/**
 * Copyright (c) 2008 - 2015 Smaxe Ltd (www.smaxe.com).
 * All rights reserved.
 */
package com.smaxe.uv.media.sdk.example.player;

import com.smaxe.uv.media.core.AudioFrame;
import com.smaxe.uv.media.core.MediaTrackInfo;
import com.smaxe.uv.media.core.VideoFrame;
import com.smaxe.uv.media.java.audio.JavaAudioPlayback;
import com.smaxe.uv.media.java.swing.JVideoScreen;
import com.smaxe.uv.media.player.IMediaPlayerModel;
import com.smaxe.uv.media.player.JMediaPlayer;
import com.smaxe.uv.media.sdk.IMediaSdk;
import com.smaxe.uv.media.sdk.MediaSdkFactory;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.DefaultBoundedRangeModel;
import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JSlider;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;
import javax.swing.event.MouseInputListener;

/**
 * <code>MediaPlayer</code> - media player.
 * 
 * @author Andrei Sochirca
 * @see <a href="http://www.smaxe.com/product.jsf?id=juv-media-sdk" target="_blank">JUV Media SDK</a>
 */
public final class MediaPlayer extends JFrame
{
    /**
     * Entry point.
     * 
     * @param args
     * @throws Exception if an exception occurred
     */
    public static void main(final String[] args) throws Exception
    {
        // NOTE:
        // Evaluation version has some limitations:
        // - video decoder watermarks frames (and blacks it sometimes)
        
        
        final IMediaSdk mediaSdk = MediaSdkFactory.getMediaSdk();
        
        
        final JFrame frame = new JFrame();
        
        {
            final JPanel contentPane = (JPanel) frame.getContentPane();
            
            final JPanel content = new JPanel(new FlowLayout());
            
            contentPane.setLayout(new BorderLayout());
            contentPane.add(content, BorderLayout.NORTH);
            
            
            final JTextField url = new JTextField(32);
            
            content.add(url);
            content.add(createButton(new AbstractAction()
            {
                private final static long serialVersionUID = 1L;
                
                public void actionPerformed(final ActionEvent e)
                {
                    final JFileChooser fileChooser = new JFileChooser();
                    
                    final int rv = fileChooser.showOpenDialog(frame);
                    if (rv != JFileChooser.APPROVE_OPTION) return;
                    
                    final File file = fileChooser.getSelectedFile();
                    
                    url.setText(file.getAbsolutePath());
                }
            }, "Open"));
            content.add(createButton(new AbstractAction()
            {
                private final static long serialVersionUID = 1L;
                
                public void actionPerformed(final ActionEvent e)
                {
                    final String path = url.getText();
                    
                    new Thread(new Runnable()
                    {
                        public void run()
                        {
                            try
                            {
                                final IMediaPlayerModel model = mediaSdk.createMediaPlayerModelSdk().create(path);
                                if (model == null) throw new Exception("Media file format is not supported");
                                
                                SwingUtilities.invokeLater(new Runnable()
                                {
                                    public void run()
                                    {
                                        MediaPlayer player = new MediaPlayer(model);
                                        
                                        setAlwaysOnTop(player, true);
                                        
                                        player.play();
                                    }
                                });
                            }
                            catch (Exception e)
                            {
                                e.printStackTrace();
                            }
                        }
                    }).start();
                }
            }, "Play"));
        }
        
        
        frame.addWindowListener(new WindowAdapter()
        {
            @Override
            public void windowClosing(WindowEvent e)
            {
                System.exit(0);
            }
        });
        
        frame.setTitle("Java Media Player");
        frame.setVisible(true);
        frame.pack();
    }
    
    /**
     * Invokes {@link Window#setAlwaysOnTop(boolean)} wrapping it in the try/catch.
     * 
     * @param window
     * @param alwaysOnTop
     */
    public static void setAlwaysOnTop(final Window window, final boolean alwaysOnTop)
    {
        if (!alwaysOnTop) return;
        
        try
        {
            window.setAlwaysOnTop(alwaysOnTop);
        }
        catch (Exception e) {/*ignore*/}
    }
    
    
    private final static long serialVersionUID = 1L;
    
    /**
     * <code>IMediaControl</code> - media control interface.
     */
    private interface IMediaControl
    {
        /**
         * Notifies about playback speed change.
         * 
         * @param speed new playback speed
         */
        public void onPlaybackSpeed(final float speed);
        
        /**
         * Notifies about a new {@link AudioFrame <tt>frame</tt>}.
         * 
         * @param frame
         */
        public void onAudioFrame(final AudioFrame frame);
        
        /**
         * Notifies about a new {@link VideoFrame <tt>frame</tt>}.
         * 
         * @param frame
         */
        public void onVideoFrame(final VideoFrame frame);
    }
    
    /**
     * <code>MediaPlayerListener</code> - media player listener.
     */
    private final class MediaPlayerListener extends JMediaPlayer.ListenerAdapter implements JMediaPlayer.IListener
    {
        // fields
        private VideoFrame lastFrame = new VideoFrame(-1, -1, null);
        
        /**
         * Constructor.
         */
        public MediaPlayerListener()
        {
        }
        
        // JMediaPlayer.IListener implementation
        
        @Override
        public void onAudioFrame(final AudioFrame frame)
        {
            if (control != null) control.onAudioFrame(frame);
        }
        
        @Override
        public void onVideoFrame(final VideoFrame frame, final int[] frameCropOffset)
        {
            if (control != null) control.onVideoFrame(frame);
            
            SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    if (hideFrameBorders)
                    {
                        videoScreen.setFrameCropOffset(frameCropOffset);
                    }
                    else
                    {
                        videoScreen.setFrameCropOffset((Insets) null);
                    }
                    
                    videoScreen.setSubtitle(frameShowTimeInSubtitle ? timeToString(0 + frame.timestamp, true) : (String) null);
                    videoScreen.setFrame(frame);
                    
                    if (frame.width != lastFrame.width || frame.height != lastFrame.height)
                    {
                        // if frame is too large -> downscale the frame
                        if (frame.width > 1200)
                        {
                            new SetFrameScaleAction(0.5f).actionPerformed(null);                            
                        }
                        else
                        {
                            new SetFrameScaleAction(1f).actionPerformed(null);                            
                        }
                        
                        pack();
                        
                        if (!MediaPlayer.this.isVisible())
                        {
                            if (infoDialog != null) infoDialog.setVisible(false);
                            MediaPlayer.this.setLocation(dialogLocation);
                            MediaPlayer.this.setVisible(true);
                        }
                    }
                    
                    lastFrame = frame;
                }
            });
        }
    }
    
    /**
     * <code>MediaInfoPanel</code> - media info panel.
     */
    private final class MediaInfoPanel extends JPanel
    {
        private final static long serialVersionUID = -841000965553781615L;
        
        // fields
        private JTabbedPane tabbedPane;
        
        /**
         * Constructor.
         */
        public MediaInfoPanel()
        {
            super(new BorderLayout());
            
            tabbedPane = new JTabbedPane(JTabbedPane.TOP);
            
            tabbedPane.setFocusable(false);
            tabbedPane.addTab("Tracks & Codecs", new JPanel());
            tabbedPane.addTab("Statistics", new JPanel());
            
            add(tabbedPane);
            setPreferredSize(new Dimension(640, 480));
        }
    }
    
    /**
     * <code>MediaControl</code> - {@link IMediaControl} for local media streams.
     */
    private final class MediaControl extends JPanel implements IMediaControl
    {
        private final static long serialVersionUID = 7943709151003379497L;
        
        // fields
        // configuration
        private boolean millisecondsPrecision = false;
        // state
        private boolean audioOnlyStream = false;
        private long lastTimestamp = -9999;
        // ui
        private JSlider timer;
        private JLabel speed;
        private JLabel status;
        
        /**
         * Constructor.
         * 
         * @param model
         * @param tracks
         */
        public MediaControl(final IMediaPlayerModel model, final MediaTrackInfo[] tracks)
        {
            setPreferredSize(new Dimension(-1, 64));
            createComponents();
            
            audioOnlyStream = tracks.length == 1 && MediaTrackInfo.findFirstTrack(tracks, MediaTrackInfo.AUDIO) != null;
            
            timer.setModel(new DefaultBoundedRangeModel(0, 100, 0, Math.max((int) tracks[0].duration, 100)));
            
            timer.setMajorTickSpacing(-1);
            timer.setMinorTickSpacing(-1);
            timer.setSnapToTicks(true);
            
            timer.addMouseListener(new MouseAdapter()
            {
                @Override
                public void mouseReleased(MouseEvent e)
                {
                    final int v = e.getX() * timer.getMaximum() / timer.getWidth();
                    
                    timer.setValue(v);
                    
                    seek(v);
                }
            });
        }
        
        // IMediaControl implementation
        
        public void onPlaybackSpeed(final float playbackSpeed)
        {
            SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    speed.setText(playbackSpeed + "x ");
                }
            });
        }
        
        public void onAudioFrame(final AudioFrame frame)
        {
            if (!audioOnlyStream) return;
            
            setTimestamp(frame.timestamp);
        }
        
        public void onVideoFrame(final VideoFrame frame)
        {
            setTimestamp(frame.timestamp);
        }
        
        // inner use methods
        /**
         * Creates components.
         */
        private void createComponents()
        {
            timer = new JSlider();
            status = new JLabel();
            speed = new JLabel("1x ");
            
            timer.setFocusable(false);
            
            JPanel statusPanel = new JPanel(new FlowLayout());
            
            statusPanel.add(speed);
            statusPanel.add(status);
            
            JPanel infoPanel = new JPanel(new BorderLayout());
            
            infoPanel.setBorder(new EmptyBorder(8, 8, 8, 8));
            infoPanel.add(createButton(new ShowMediaInfoAction(), "Info"), BorderLayout.WEST);
            infoPanel.add(statusPanel, BorderLayout.EAST);
            
            setLayout(new BorderLayout());
            
            add(timer, BorderLayout.NORTH);
            add(infoPanel, BorderLayout.SOUTH);
        }
        
        /**
         * Sets timestamp.
         * 
         * @param timestamp
         */
        private void setTimestamp(final long timestamp)
        {
            if (!millisecondsPrecision)
            {
                final long diff = timestamp - lastTimestamp;
                if (diff >= 0 && diff < 1000) return;
                
                lastTimestamp = timestamp;
            }
            
            SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    status.setText(timeToString(timestamp, millisecondsPrecision));
                    
                    timer.setValue((int) timestamp);
                }
            });
        }
    }
    
    /**
     * <code>LiveMediaControl</code> - {@link IMediaControl} for live media streams.
     */
    private final class LiveMediaControl extends JPanel implements IMediaControl
    {
        private final static long serialVersionUID = -4719693193945752963L;
        
        // fields
        // ui
        private JLabel status;
        
        /**
         * Constructor.
         */
        public LiveMediaControl()
        {
            setPreferredSize(new Dimension(-1, 48));
            createComponents();
        }
        
        // IMediaControl implementation
        
        public void onPlaybackSpeed(final float speed)
        {
        }
        
        public void onAudioFrame(final AudioFrame frame)
        {
        }
        
        public void onVideoFrame(final VideoFrame frame)
        {
            setTimestamp(frame.timestamp);
        }
        
        // inner use methods
        /**
         * Creates components.
         */
        private void createComponents()
        {
            setLayout(new BorderLayout());
            
            JPanel info = new JPanel(new BorderLayout());
            
            info.setBorder(new EmptyBorder(8, 8, 8, 8));
            info.add(createButton(new ShowMediaInfoAction(), "Info"), BorderLayout.WEST);
            info.add(status = new JLabel(), BorderLayout.EAST);
            
            add(info, BorderLayout.SOUTH);
        }
        
        /**
         * Sets timestamp.
         * 
         * @param timestamp
         */
        private void setTimestamp(final long timestamp)
        {
            SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    status.setText(timeToString(timestamp, false));
                }
            });
        }
    }
    
    /**
     * <code>VideoScreenMouseListener</code> - video screen mouse listener.
     */
    private final class VideoScreenMouseListener extends MouseAdapter implements MouseInputListener
    {
        // fields
        private Point mousePressedPosition = null;
        private long lastFrameLocationChangeTimestamp = 0;
        
        /**
         * Constructor.
         */
        public VideoScreenMouseListener()
        {
        }
        
        // MouseListener implementation
        
        @Override
        public void mousePressed(final MouseEvent e)
        {
            mousePressedPosition = getLocationOnScreen(e);
        }
        
        // MouseMotionListener implementation
        
        @Override
        public void mouseDragged(final MouseEvent e)
        {
            if (!SwingUtilities.isLeftMouseButton(e)) return;
            
            final long ctime = System.currentTimeMillis();
            final long diff = ctime - lastFrameLocationChangeTimestamp;
            if (diff >= 0 && diff < 50) return;
            
            final Point mouseDraggedPosition = getLocationOnScreen(e);
            
            int dx = mouseDraggedPosition.x - mousePressedPosition.x;
            int dy = mouseDraggedPosition.y - mousePressedPosition.y;
            
            if (dx == 0 && dy == 0) return;
            
            Rectangle bounds = MediaPlayer.this.getBounds();
            
            MediaPlayer.this.setBounds(bounds.x + dx, bounds.y + dy, bounds.width, bounds.height);
            
            lastFrameLocationChangeTimestamp = ctime;
            mousePressedPosition = mouseDraggedPosition;
        }
        
        // inner use methods
        /**
         * Returns {@link MouseEvent <tt>e</tt>} location on screen.
         * 
         * @param e
         * @return location on screen
         */
        private Point getLocationOnScreen(final MouseEvent e)
        {
            return e.getLocationOnScreen();
        }
    }
    
    /**
     * <code>SetFrameAutoScaleAction</code> - set 'Frame Auto Scale' action.
     */
    private final class SetFrameAutoScaleAction extends AbstractAction
    {
        private final static long serialVersionUID = -648502523570195340L;
        
        /**
         * Constructor.
         */
        public SetFrameAutoScaleAction()
        {
        }
        
        // Action implementation
        
        public void actionPerformed(final ActionEvent e)
        {
            videoScreen.setAutoScaleOnResize(!videoScreen.getAutoScaleOnResize());
        }
    }
    
    /**
     * <code>SetFrameScaleAction</code> - set 'Frame Scale' action.
     */
    private final class SetFrameScaleAction extends AbstractAction
    {
        private final static long serialVersionUID = -648502523570195340L;
        
        // fields
        private final float scale;
        
        /**
         * Constructor.
         * 
         * @param scale
         */
        public SetFrameScaleAction(final float scale)
        {
            this.scale = scale;
        }
        
        // Action implementation
        
        public void actionPerformed(final ActionEvent e)
        {
            if (fullscreenDialog != null)
            {
                getContentPane().add(videoScreen, BorderLayout.CENTER);
                
                dispose(fullscreenDialog);
                
                fullscreenDialog = null;
            }
            
            videoScreen.scale(scale);
            
            pack();
        }
    }
    
    /**
     * <code>SetFullScreenAction</code> - set 'Full Screen' action.
     */
    private final class SetFullScreenAction extends AbstractAction
    {
        private final static long serialVersionUID = -648502523570125340L;
        
        /**
         * Constructor.
         */
        public SetFullScreenAction()
        {
        }
        
        // Action implementation
        
        public void actionPerformed(final ActionEvent e)
        {
            final Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
            
            JDialog dialog = new JDialog();
            
            setAlwaysOnTop(dialog, true);
            dialog.setUndecorated(true);
            dialog.setPreferredSize(dim);
            dialog.setSize(dim);
            dialog.setLocation(0, 0);
            dialog.getContentPane().setLayout(new BorderLayout());
            dialog.getContentPane().add(videoScreen, BorderLayout.CENTER);
            
            dialog.addKeyListener(new KeyAdapter()
            {
                @Override
                public void keyReleased(final KeyEvent e)
                {
                    switch (e.getKeyCode())
                    {
                        case KeyEvent.VK_ESCAPE:
                        case KeyEvent.VK_ENTER:
                        {
                            new SetFrameScaleAction(1f).actionPerformed(null);
                        } break;
                    }
                }
            });
            
            fullscreenDialog = dialog;
            
            fullscreenDialog.setVisible(true);
        }
    }
    
    /**
     * <code>SetPlaybackSpeedAction</code> - set 'Playback Speed' action.
     */
    private final class SetPlaybackSpeedAction extends AbstractAction
    {
        private final static long serialVersionUID = -3217219113209740745L;
        
        // fields
        private final float speed;
        
        /**
         * Constructor.
         * 
         * @param speed
         */
        public SetPlaybackSpeedAction(final float speed)
        {
            this.speed = speed;
        }
        
        // Action implementation
        
        public void actionPerformed(final ActionEvent e)
        {
            player.setPlaybackSpeed(speed);
            
            if (player.getPlaybackSpeed() == speed)
            {
                control.onPlaybackSpeed(playbackSpeed = speed);
            }
        }
    }
    
    /**
     * <code>ShowMediaInfoAction</code> - 'Show Media Info' action.
     */
    private final class ShowMediaInfoAction extends AbstractAction
    {
        private final static long serialVersionUID = -7483331948036595604L;
        
        /**
         * Constructor.
         * 
         * @param speed
         */
        public ShowMediaInfoAction()
        {
        }
        
        // Action implementation
        
        public void actionPerformed(final ActionEvent e)
        {
            showMediaInfoDialog();
        }
    };
    
    // fields
    // control
    private IMediaControl control;
    // player
    private JMediaPlayer player;
    // ui
    private JVideoScreen videoScreen;
    private MediaInfoPanel infoPanel;
    // dialogs
    private JDialog fullscreenDialog;
    private JDialog infoDialog;
    // state
    private Point dialogLocation = new Point(0, 0);
    private boolean hideFrameBorders = true;
    private float playbackSpeed = 1f;
    private boolean frameShowTimeInSubtitle = false;
    
    /**
     * Constructor.
     * 
     * @param mediaPlayerModel media player model
     */
    public MediaPlayer(final IMediaPlayerModel mediaPlayerModel)
    {
        player = JMediaPlayer.create(mediaPlayerModel);
        
        player.setAudioPlayback(new JavaAudioPlayback());
        
        player.setListener(new MediaPlayerListener());
        
        addWindowListener(new WindowAdapter()
        {
            @Override
            public void windowClosing(WindowEvent e)
            {
                player.stopPlayback();
            }
        });
        
        infoPanel = new MediaInfoPanel();
        
        videoScreen = new JVideoScreen();
        
        videoScreen.setBackground(Color.BLACK);
        
        videoScreen.addMouseListener(new MouseAdapter()
        {
            @Override
            public void mouseClicked(final MouseEvent e)
            {
                final int clickCount = e.getClickCount();
                final Object source = e.getSource();
                
                switch (clickCount)
                {
                    case 1:
                    {
                        if (SwingUtilities.isRightMouseButton(e))
                        {
                            JPopupMenu popup = new JPopupMenu();
                            
                            if (player.isPaused())
                            {
                                popup.add(new AbstractAction("Resume")
                                {
                                    private final static long serialVersionUID = 0L;
                                    
                                    public void actionPerformed(ActionEvent e)
                                    {
                                        player.pause(false);
                                    }
                                });
                            }
                            else
                            {
                                popup.add(new AbstractAction("Pause")
                                {
                                    private final static long serialVersionUID = 0L;
                                    
                                    public void actionPerformed(ActionEvent e)
                                    {
                                        player.pause(true);
                                    }
                                });
                            }
                            
                            popup.addSeparator();
                            
                            JMenu scaleMenu = new JMenu("Scale");
                            
                            double scale = videoScreen.getScaleFactor();
                            
                            scaleMenu.add(createCheckBoxMenuItem(new SetFrameAutoScaleAction(), "Auto", videoScreen.getAutoScaleOnResize()));
                            scaleMenu.addSeparator();
                            scaleMenu.add(createCheckBoxMenuItem(new SetFrameScaleAction(0.5f), "0.5x", scale == 0.5f));
                            scaleMenu.add(createCheckBoxMenuItem(new SetFrameScaleAction(1f), "1x", scale == 1f));
                            scaleMenu.add(createCheckBoxMenuItem(new SetFrameScaleAction(2f), "2x", scale == 2f));
                            scaleMenu.add(createCheckBoxMenuItem(new SetFrameScaleAction(4f), "4x", scale == 4f));
                            scaleMenu.addSeparator();
                            scaleMenu.add(createCheckBoxMenuItem(new SetFullScreenAction(), "Full", false));
                            
                            popup.add(scaleMenu);
                            
                            popup.add(createCheckBoxMenuItem(new AbstractAction()
                            {
                                private final static long serialVersionUID = 0L;
                                
                                public void actionPerformed(ActionEvent e)
                                {
                                    videoScreen.mirror(!videoScreen.getMirrorFlag());
                                }
                            }, "Mirror", videoScreen.getMirrorFlag()));
                            
                            popup.add(createCheckBoxMenuItem(new AbstractAction()
                            {
                                private final static long serialVersionUID = 0L;
                                
                                public void actionPerformed(ActionEvent e)
                                {
                                    videoScreen.flip(!videoScreen.getFlipFlag());
                                }
                            }, "Flip", videoScreen.getFlipFlag()));
                            
                            if (!player.isLive())
                            {
                                JMenu playbackSpeedMenu = new JMenu("Speed");
                                
                                playbackSpeedMenu.add(createCheckBoxMenuItem(new SetPlaybackSpeedAction(0.5f), "0.5x", playbackSpeed == 0.5f));
                                playbackSpeedMenu.add(createCheckBoxMenuItem(new SetPlaybackSpeedAction(0.75f), "0.75x", playbackSpeed == 0.75f));
                                playbackSpeedMenu.add(createCheckBoxMenuItem(new SetPlaybackSpeedAction(1f), "1x", playbackSpeed == 1f));
                                playbackSpeedMenu.add(createCheckBoxMenuItem(new SetPlaybackSpeedAction(1.25f), "1.25x", playbackSpeed == 1.25f));
                                playbackSpeedMenu.add(createCheckBoxMenuItem(new SetPlaybackSpeedAction(1.5f), "1.5x", playbackSpeed == 1.5f));
                                playbackSpeedMenu.add(createCheckBoxMenuItem(new SetPlaybackSpeedAction(2f), "2x", playbackSpeed == 2f));
                                
                                popup.addSeparator();
                                popup.add(playbackSpeedMenu);
                            }
                            
                            popup.show((Component) source, e.getX(), e.getY());
                        }
                    } break;
                }
            }
        });
        
        final VideoScreenMouseListener mouseListener = new VideoScreenMouseListener();
        
        videoScreen.addMouseListener(mouseListener);
        videoScreen.addMouseMotionListener(mouseListener);
        
        addKeyListener(new KeyAdapter()
        {
            @Override
            public void keyReleased(final KeyEvent e)
            {
                switch (e.getKeyCode())
                {
                    case KeyEvent.VK_ENTER:
                    {
                        new SetFullScreenAction().actionPerformed(null);
                    } break;
                    case KeyEvent.VK_UP:
                    {
                        if (e.isControlDown())
                        {
                            new SetPlaybackSpeedAction(playbackSpeed * 2).actionPerformed(null);
                        }
                    } break;
                    case KeyEvent.VK_DOWN:
                    {
                        if (e.isControlDown())
                        {
                            new SetPlaybackSpeedAction(playbackSpeed / 2).actionPerformed(null);
                        }
                    } break;
                }
            }
        });
        
        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(videoScreen, BorderLayout.CENTER);
        
        setTitle(mediaPlayerModel.getSource());
        
        showMediaInfoDialog();
    }
    
    /**
     * Starts the playback.
     */
    public void play()
    {
        final IMediaPlayerModel model = player.getModel();
        
        model.addListener(new IMediaPlayerModel.ListenerAdapter()
        {
            @Override
            public void onTracks(final boolean live, final MediaTrackInfo[] tracks, final MediaTrackInfo[] all)
            {
                if (tracks.length == 1 && tracks[0].type == MediaTrackInfo.AUDIO)
                {
                    SwingUtilities.invokeLater(new Runnable()
                    {
                        public void run()
                        {
                            control = live ? new LiveMediaControl() : new MediaControl(model, tracks);
                            
                            getContentPane().add((JPanel) control, BorderLayout.SOUTH);
                            
                            pack();
                            
                            if (!MediaPlayer.this.isVisible())
                            {
                                if (infoDialog != null) infoDialog.setVisible(false);
                                MediaPlayer.this.setLocation(dialogLocation);
                                MediaPlayer.this.setVisible(true);
                            }
                        }
                    });
                }
                else
                {
                    SwingUtilities.invokeLater(new Runnable()
                    {
                        public void run()
                        {
                            control = live ? new LiveMediaControl() : new MediaControl(model, tracks);
                            
                            getContentPane().add((JPanel) control, BorderLayout.SOUTH);
                            
                            pack();
                        }
                    });
                }
            }
        });
        
        player.startPlayback();
    }
    
    /**
     * Pauses/resumes the playback.
     * 
     * @param pause
     */
    public void pause(final boolean pause)
    {
        player.pause(pause);
    }
    
    /**
     * Seeks specified position.
     * 
     * @param timestamp
     */
    public void seek(final long timestamp)
    {
        player.seek(timestamp);
    }
    
    // inner use methods
    /**
     * Shows media info dialog.
     */
    private void showMediaInfoDialog()
    {
        if (infoDialog != null)
        {
            infoDialog.setVisible(true);
            return;
        }
        
        final JDialog dialog = new JDialog();
        
        dialog.getContentPane().setLayout(new BorderLayout());
        dialog.getContentPane().add(infoPanel, BorderLayout.CENTER);
        
        setAlwaysOnTop(dialog, true);
        dialog.setLocation(dialogLocation);
        dialog.setTitle(player.getModel().getSource() + " - Info");
        
        dialog.addComponentListener(new ComponentAdapter()
        {
            @Override
            public void componentMoved(ComponentEvent e)
            {
                dialogLocation = dialog.getLocation();
            }
        });
        
        dialog.addWindowListener(new WindowAdapter()
        {
            @Override
            public void windowClosing(WindowEvent e)
            {
                dispose(dialog);
                
                infoDialog = null;
            }
        });
        
        infoDialog = dialog;
        
        dialog.pack();
        dialog.setVisible(true);
    }
    
    // TimeUtils methods
    /**
     * <code>HOUR</code>
     */
    private final static long HOUR = 1 * 60 * 60 * 1000;
    /**
     * <code>MINUTE</code>
     */
    private final static long MINUTE = 1 * 60 * 1000;
    /**
     * <code>SECOND</code>
     */
    private final static long SECOND = 1 * 1000;
    
    /**
     * Returns <code>time</code> string in 'hh:mm:ss.zzz' or 'hh:mm:ss' format
     * 
     * @param time time (in milliseconds)
     * @param milliseconds <code>true</code> to use 'hh:mm:ss.zzz' format
     * @return string representation of the time
     */
    private static String timeToString(final long time, final boolean milliseconds)
    {
        final int seconds = (int) ((time % MINUTE) / SECOND);
        final int minutes = (int) ((time % HOUR) / MINUTE);
        final int hours = (int) (time / HOUR);
        
        if (hours > 0)
        {
            if (milliseconds)
            {
                return new String(new char[]
                {
                        (char) ('0' + (hours / 10)), (char) ('0' + (hours % 10)),
                        ':',
                        (char) ('0' + (minutes / 10)), (char) ('0' + (minutes % 10)),
                        ':',
                        (char) ('0' + (seconds / 10)), (char) ('0' + (seconds % 10)),
                        '.',
                        (char) ('0' + ((time % 1000) / 100)), (char) ('0' + ((time % 100) / 10)), (char) ('0' + (time % 10))
                });
            }
            else
            {
                return new String(new char[]
                {
                        (char) ('0' + (hours / 10)), (char) ('0' + (hours % 10)),
                        ':',
                        (char) ('0' + (minutes / 10)), (char) ('0' + (minutes % 10)),
                        ':',
                        (char) ('0' + (seconds / 10)), (char) ('0' + (seconds % 10)),
                });
            }
        }
        else
        {
            if (milliseconds)
            {
                return new String(new char[]
                {
                        (char) ('0' + (minutes / 10)), (char) ('0' + (minutes % 10)),
                        ':',
                        (char) ('0' + (seconds / 10)), (char) ('0' + (seconds % 10)),
                        '.',
                        (char) ('0' + ((time % 1000) / 100)), (char) ('0' + ((time % 100) / 10)), (char) ('0' + (time % 10))
                });
            }
            else
            {
                return new String(new char[]
                {
                        (char) ('0' + (minutes / 10)), (char) ('0' + (minutes % 10)),
                        ':',
                        (char) ('0' + (seconds / 10)), (char) ('0' + (seconds % 10)),
                });
            }
        }
    }
    
    // UIUtils methods
    /**
     * @param action
     * @param text
     * @return button
     */
    private static JButton createButton(final Action action, final String text)
    {
        JButton button = new JButton(action);
        
        button.setFocusable(false);
        button.setText(text);
        
        return button;
    }
    
    /**
     * @param action
     * @param text
     * @param selected
     * @return menu item
     */
    private static JCheckBoxMenuItem createCheckBoxMenuItem(final Action action, final String text,
            final boolean selected)
    {
        JCheckBoxMenuItem item = new JCheckBoxMenuItem(action);
        
        item.setText(text);
        item.setSelected(selected);
        
        return item;
    }
    
    /**
     * Disposes the {@link Window <tt>window</tt>}.
     * 
     * @param window
     */
    private static void dispose(final Window window)
    {
        if (window == null) return;
        
        window.setVisible(false);
        window.removeAll();
        window.dispose();
    }
}